import os
import re
import uuid
from typing import Any, Dict, List

from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.errors import UserError

RETRYABLE_CODES = [
    status_code_pb2.MODEL_DEPLOYING,
    status_code_pb2.MODEL_LOADING,
    status_code_pb2.MODEL_BUSY_PLEASE_RETRY,
]

DEFAULT_CONFIG = f'{os.environ["HOME"]}/.config/clarifai/config'


def status_is_retryable(status_code: int) -> bool:
    """Check if a status code is retryable."""
    return status_code in RETRYABLE_CODES


class Chunker:
    """Split an input sequence into small chunks."""

    def __init__(self, seq: List, size: int) -> None:
        self.seq = seq
        self.size = size

    def chunk(self) -> List[List]:
        """Chunk input sequence."""
        return [self.seq[pos : pos + self.size] for pos in range(0, len(self.seq), self.size)]


class BackoffIterator:
    """Iterator that returns a sequence of backoff values."""

    def __init__(self, count=0):
        self.count = count

    def __iter__(self):
        return self

    def __next__(self):
        self.count += 1
        return 0.1 * (1.3**self.count)


def get_from_dict_or_env(key: str, env_key: str, **data) -> str:
    """Get a value from a dictionary or an environment variable."""
    if key in data and data[key]:
        return data[key]
    else:
        return get_from_env(key, env_key)


def get_from_env(key: str, env_key: str) -> str:
    """Get a value from a dictionary or an environment variable."""
    if env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]
    else:
        raise UserError(
            f"Did not find `{key}`, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f"  `{key}` as a named parameter."
        )


def concept_relations_accumulation(
    relations_dict: Dict[str, Any], subject_concept: str, object_concept: str, predicate: str
) -> Dict[str, Any]:
    """Append the concept relation to relations dict based on its predicate.

    Args:
        relations_dict (dict): A dict of concept relations info.
    """
    if predicate == 'hyponym':
        if object_concept in relations_dict:
            relations_dict[object_concept].append(subject_concept)
        else:
            relations_dict[object_concept] = [subject_concept]
    elif predicate == 'hypernym':
        if subject_concept in relations_dict:
            relations_dict[subject_concept].append(object_concept)
        else:
            relations_dict[subject_concept] = [object_concept]
    else:
        relations_dict[object_concept] = []
        relations_dict[subject_concept] = []
    return relations_dict


def get_uuid(val: int) -> str:
    """Generates a UUID."""
    return uuid.uuid4().hex[:val]


def clean_input_id(input_id: str) -> str:
    """Clean input_id string into a valid input ID"""
    input_id = re.sub('[., /]+', '_', input_id)
    input_id = re.sub('[_]+', '_', input_id)
    input_id = re.sub('[-]+', '-', input_id)
    input_id = input_id.lower().strip('_-')
    input_id = re.sub('[^a-z0-9-_]+', '', input_id)
    return input_id



import concurrent.futures

import fsspec

from clarifai.utils.logging import logger


def download_input(input):
    _download_input_data(input.data)
    if input.data.parts:
        for i in range(len(input.data.parts)):
            _download_input_data(input.data.parts[i].data)


def _download_input_data(input_data):
    """
    This function will download any urls that are not already bytes.
    """
    if input_data.image.url and not input_data.image.base64:
        # Download the image
        with fsspec.open(input_data.image.url, 'rb') as f:
            input_data.image.base64 = f.read()
    if input_data.video.url and not input_data.video.base64:
        # Download the video
        with fsspec.open(input_data.video.url, 'rb') as f:
            input_data.video.base64 = f.read()
    if input_data.audio.url and not input_data.audio.base64:
        # Download the audio
        with fsspec.open(input_data.audio.url, 'rb') as f:
            input_data.audio.base64 = f.read()
    if input_data.text.url and not input_data.text.raw:
        # Download the text
        with fsspec.open(input_data.text.url, 'r') as f:
            input_data.text.raw = f.read()


def ensure_urls_downloaded(request, max_threads=128):
    """
    This function will download any urls that are not already bytes and parallelize with a thread pool.
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        futures = []
        for input in request.inputs:
            futures.append(executor.submit(download_input, input))
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception as e:
                logger.exception(f"Error downloading input: {e}")



import inspect
import itertools
import logging
import os
import traceback
from abc import ABC
from collections import abc
from typing import Any, Dict, Iterator, List
from unittest.mock import MagicMock

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2

from clarifai.runners.utils import data_types
from clarifai.runners.utils.data_utils import DataConverter
from clarifai.runners.utils.method_signatures import (
    build_function_signature,
    deserialize,
    get_stream_from_signature,
    serialize,
    signatures_to_json,
)

_METHOD_INFO_ATTR = '_cf_method_info'

_RAISE_EXCEPTIONS = os.getenv("RAISE_EXCEPTIONS", "false").lower() in ("true", "1")


class ModelClass(ABC):
    '''
    Base class for model classes that can be run as a service.

    Define predict, generate, or stream methods using the @ModelClass.method decorator.

    Example:

      from clarifai.runners.model_class import ModelClass
      from clarifai.runners.utils.data_types import NamedFields
      from typing import List, Iterator

      class MyModel(ModelClass):

        @ModelClass.method
        def predict(self, x: str, y: int) -> List[str]:
          return [x] * y

        @ModelClass.method
        def generate(self, x: str, y: int) -> Iterator[str]:
          for i in range(y):
            yield x + str(i)

        @ModelClass.method
        def stream(self, input_stream: Iterator[NamedFields(x=str, y=int)]) -> Iterator[str]:
          for item in input_stream:
            yield item.x + ' ' + str(item.y)
    '''

    @staticmethod
    def method(func):
        setattr(func, _METHOD_INFO_ATTR, _MethodInfo(func))
        return func

    def set_output_context(self, prompt_tokens=None, completion_tokens=None):
        """This is used to set the prompt and completion tokens in the Output proto"""
        self._prompt_tokens = prompt_tokens
        self._completion_tokens = completion_tokens

    def load_model(self):
        """Load the model."""

    def _handle_get_signatures_request(self) -> service_pb2.MultiOutputResponse:
        methods = self._get_method_info()
        signatures = {method.name: method.signature for method in methods.values()}
        resp = service_pb2.MultiOutputResponse(
            status=status_pb2.Status(code=status_code_pb2.SUCCESS)
        )
        output = resp.outputs.add()
        output.status.code = status_code_pb2.SUCCESS
        output.data.text.raw = signatures_to_json(signatures)
        return resp

    def _batch_predict(self, method, inputs: List[Dict[str, Any]]) -> List[Any]:
        """Batch predict method for multiple inputs."""
        outputs = []
        for input in inputs:
            output = method(**input)
            outputs.append(output)
        return outputs

    def _batch_generate(self, method, inputs: List[Dict[str, Any]]) -> Iterator[List[Any]]:
        """Batch generate method for multiple inputs."""
        generators = [method(**input) for input in inputs]
        for outputs in itertools.zip_longest(*generators):
            yield outputs

    def predict_wrapper(
        self, request: service_pb2.PostModelOutputsRequest
    ) -> service_pb2.MultiOutputResponse:
        outputs = []
        try:
            # TODO add method name field to proto
            method_name = 'predict'
            if len(request.inputs) > 0 and '_method_name' in request.inputs[0].data.metadata:
                method_name = request.inputs[0].data.metadata['_method_name']
            if (
                method_name == '_GET_SIGNATURES'
            ):  # special case to fetch signatures, TODO add endpoint for this
                return self._handle_get_signatures_request()
            if method_name not in self._get_method_info():
                raise ValueError(f"Method {method_name} not found in model class")
            method = getattr(self, method_name)
            method_info = method._cf_method_info
            signature = method_info.signature
            python_param_types = method_info.python_param_types
            for input in request.inputs:
                # check if input is in old format
                is_convert = DataConverter.is_old_format(input.data)
                if is_convert:
                    # convert to new format
                    new_data = DataConverter.convert_input_data_to_new_format(
                        input.data, signature.input_fields
                    )
                    input.data.CopyFrom(new_data)
            # convert inputs to python types
            inputs = self._convert_input_protos_to_python(
                request.inputs, signature.input_fields, python_param_types
            )
            if len(inputs) == 1:
                inputs = inputs[0]
                output = method(**inputs)
                outputs.append(
                    self._convert_output_to_proto(
                        output, signature.output_fields, convert_old_format=is_convert
                    )
                )
            else:
                outputs = self._batch_predict(method, inputs)
                outputs = [
                    self._convert_output_to_proto(
                        output, signature.output_fields, convert_old_format=is_convert
                    )
                    for output in outputs
                ]

            return service_pb2.MultiOutputResponse(
                outputs=outputs, status=status_pb2.Status(code=status_code_pb2.SUCCESS)
            )
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            logging.exception("Error in predict")
            return service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.FAILURE,
                    details=str(e),
                    stack_trace=traceback.format_exc().split('\n'),
                )
            )

    def generate_wrapper(
        self, request: service_pb2.PostModelOutputsRequest
    ) -> Iterator[service_pb2.MultiOutputResponse]:
        try:
            method_name = 'generate'
            if len(request.inputs) > 0 and '_method_name' in request.inputs[0].data.metadata:
                method_name = request.inputs[0].data.metadata['_method_name']
            method = getattr(self, method_name)
            method_info = method._cf_method_info
            signature = method_info.signature
            python_param_types = method_info.python_param_types
            for input in request.inputs:
                # check if input is in old format
                is_convert = DataConverter.is_old_format(input.data)
                if is_convert:
                    # convert to new format
                    new_data = DataConverter.convert_input_data_to_new_format(
                        input.data, signature.input_fields
                    )
                    input.data.CopyFrom(new_data)
            inputs = self._convert_input_protos_to_python(
                request.inputs, signature.input_fields, python_param_types
            )
            if len(inputs) == 1:
                inputs = inputs[0]
                for output in method(**inputs):
                    resp = service_pb2.MultiOutputResponse()
                    self._convert_output_to_proto(
                        output,
                        signature.output_fields,
                        proto=resp.outputs.add(),
                        convert_old_format=is_convert,
                    )
                    resp.status.code = status_code_pb2.SUCCESS
                    yield resp
            else:
                for outputs in self._batch_generate(method, inputs):
                    resp = service_pb2.MultiOutputResponse()
                    for output in outputs:
                        self._convert_output_to_proto(
                            output,
                            signature.output_fields,
                            proto=resp.outputs.add(),
                            convert_old_format=is_convert,
                        )
                    resp.status.code = status_code_pb2.SUCCESS
                    yield resp
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            logging.exception("Error in generate")
            yield service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.FAILURE,
                    details=str(e),
                    stack_trace=traceback.format_exc().split('\n'),
                )
            )

    def stream_wrapper(
        self, request_iterator: Iterator[service_pb2.PostModelOutputsRequest]
    ) -> Iterator[service_pb2.MultiOutputResponse]:
        try:
            request = next(request_iterator)  # get first request to determine method
            assert len(request.inputs) == 1, "Streaming requires exactly one input"

            method_name = 'stream'
            if len(request.inputs) > 0 and '_method_name' in request.inputs[0].data.metadata:
                method_name = request.inputs[0].data.metadata['_method_name']
            method = getattr(self, method_name)
            method_info = method._cf_method_info
            signature = method_info.signature
            python_param_types = method_info.python_param_types

            # find the streaming vars in the signature
            stream_sig = get_stream_from_signature(signature.input_fields)
            if stream_sig is None:
                raise ValueError("Streaming method must have a Stream input")
            stream_argname = stream_sig.name

            for input in request.inputs:
                # check if input is in old format
                is_convert = DataConverter.is_old_format(input.data)
                if is_convert:
                    # convert to new format
                    new_data = DataConverter.convert_input_data_to_new_format(
                        input.data, signature.input_fields
                    )
                    input.data.CopyFrom(new_data)
            # convert all inputs for the first request, including the first stream value
            inputs = self._convert_input_protos_to_python(
                request.inputs, signature.input_fields, python_param_types
            )
            kwargs = inputs[0]

            # first streaming item
            first_item = kwargs.pop(stream_argname)

            # streaming generator
            def InputStream():
                yield first_item
                # subsequent streaming items contain only the streaming input
                for request in request_iterator:
                    item = self._convert_input_protos_to_python(
                        request.inputs, [stream_sig], python_param_types
                    )
                    item = item[0][stream_argname]
                    yield item

            # add stream generator back to the input kwargs
            kwargs[stream_argname] = InputStream()

            for output in method(**kwargs):
                resp = service_pb2.MultiOutputResponse()
                self._convert_output_to_proto(
                    output,
                    signature.output_fields,
                    proto=resp.outputs.add(),
                    convert_old_format=is_convert,
                )
                resp.status.code = status_code_pb2.SUCCESS
                yield resp
        except Exception as e:
            if _RAISE_EXCEPTIONS:
                raise
            logging.exception("Error in stream")
            yield service_pb2.MultiOutputResponse(
                status=status_pb2.Status(
                    code=status_code_pb2.FAILURE,
                    details=str(e),
                    stack_trace=traceback.format_exc().split('\n'),
                )
            )

    def _convert_input_protos_to_python(
        self,
        inputs: List[resources_pb2.Input],
        variables_signature: List[resources_pb2.ModelTypeField],
        python_param_types,
    ) -> List[Dict[str, Any]]:
        result = []
        for input in inputs:
            kwargs = deserialize(input.data, variables_signature)
            # dynamic cast to annotated types
            for k, v in kwargs.items():
                if k not in python_param_types:
                    continue

                if hasattr(python_param_types[k], "__args__") and (
                    getattr(python_param_types[k], "__origin__", None)
                    in [abc.Iterator, abc.Generator, abc.Iterable]
                ):
                    # get the type of the items in the stream
                    stream_type = python_param_types[k].__args__[0]

                    kwargs[k] = data_types.cast(v, stream_type)
                else:
                    kwargs[k] = data_types.cast(v, python_param_types[k])
            result.append(kwargs)
        return result

    def _convert_output_to_proto(
        self,
        output: Any,
        variables_signature: List[resources_pb2.ModelTypeField],
        proto=None,
        convert_old_format=False,
    ) -> resources_pb2.Output:
        if proto is None:
            proto = resources_pb2.Output()
        serialize({'return': output}, variables_signature, proto.data, is_output=True)
        if convert_old_format:
            # convert to old format
            data = DataConverter.convert_output_data_to_old_format(proto.data)
            proto.data.CopyFrom(data)
        proto.status.code = status_code_pb2.SUCCESS
        if hasattr(self, "_prompt_tokens") and self._prompt_tokens is not None:
            proto.prompt_tokens = self._prompt_tokens
        if hasattr(self, "_completion_tokens") and self._completion_tokens is not None:
            proto.completion_tokens = self._completion_tokens
        self._prompt_tokens = None
        self._completion_tokens = None
        return proto

    @classmethod
    def _register_model_methods(cls):
        # go up the class hierarchy to find all decorated methods, and add to registry of current class
        methods = {}
        for base in reversed(cls.__mro__):
            for name, method in base.__dict__.items():
                # Skip class attributes, mocked objects, and non-methods
                if not callable(method) or isinstance(method, (classmethod, staticmethod)):
                    continue
                # Skip any mocked objects or attributes
                if isinstance(method, MagicMock) or hasattr(method, '_mock_return_value'):
                    continue
                # Only include methods that have been decorated with @ModelClass.method
                method_info = getattr(method, _METHOD_INFO_ATTR, None)
                if not method_info:  # regular function, not a model method
                    continue
                methods[name] = method_info
        # check for generic predict(request) -> response, etc. methods
        # for name in ('predict', 'generate', 'stream'):
        #  if hasattr(cls, name):
        #    method = getattr(cls, name)
        #    if not hasattr(method, _METHOD_INFO_ATTR):  # not already put in registry
        #      methods[name] = _MethodInfo(method)
        # set method table for this class in the registry
        return methods

    @classmethod
    def _get_method_info(cls, func_name=None):
        if not hasattr(cls, _METHOD_INFO_ATTR):
            setattr(cls, _METHOD_INFO_ATTR, cls._register_model_methods())
        method_info = getattr(cls, _METHOD_INFO_ATTR)
        if func_name:
            return method_info[func_name]
        return method_info


class _MethodInfo:
    def __init__(self, method):
        self.name = method.__name__
        self.signature = build_function_signature(method)
        self.python_param_types = {
            p.name: p.annotation
            for p in inspect.signature(method).parameters.values()
            if p.annotation != inspect.Parameter.empty
        }
        self.python_param_types.pop('self', None)



"""Base class for creating Model Context Protocol (MCP) servers."""

import asyncio
import json
from typing import Any

from fastmcp import Client, FastMCP  # use fastmcp v2 not the built in mcp
from mcp import types
from mcp.shared.exceptions import McpError

from clarifai.runners.models.model_class import ModelClass


class MCPModelClass(ModelClass):
    """Base class for wrapping FastMCP servers as a model running in Clarfai. This handles
    all the transport between the API and the MCP server here. Simply subclass this and implement
    the get_server() method to return the FastMCP server instance. The server is then used to
    handle all the requests and responses.
    """

    def load_model(self):
        # in memory transport provided in fastmcp v2 so we can easily use the client functions.
        self.client = Client(self.get_server())

    def get_server(self) -> FastMCP:
        """Required method for each subclass to implement to return the FastMCP server to use."""
        raise NotImplementedError("Subclasses must implement get_server() method")

    @ModelClass.method
    def mcp_transport(self, msg: str) -> str:
        """The single model method to get the jsonrpc message and send it to the FastMCP server then
        return it's response.

        """

        async def send_notification(client_message: types.ClientNotification) -> None:
            async with self.client:
                # Strip the jsonrpc field since send_notification will also pass it in for some reason.
                client_message = types.ClientNotification.model_validate(
                    client_message.model_dump(
                        by_alias=True, mode="json", exclude_none=True, exclude={"jsonrpc"}
                    )
                )
                try:
                    return await self.client.session.send_notification(client_message)
                except McpError as e:
                    return types.JSONRPCError(jsonrpc="2.0", error=e.error)

        async def send_request(client_message: types.ClientRequest, id: str) -> Any:
            async with self.client:
                # Strip the jsonrpc and id fields as send_request sets them again too.
                client_message = types.ClientRequest.model_validate(
                    client_message.model_dump(
                        by_alias=True, mode="json", exclude_none=True, exclude={"jsonrpc", "id"}
                    )
                )

                result_type = None
                if isinstance(client_message.root, types.PingRequest):
                    result_type = types.EmptyResult
                elif isinstance(client_message.root, types.InitializeRequest):
                    return await self.client.session.initialize()
                elif isinstance(client_message.root, types.SetLevelRequest):
                    result_type = types.EmptyResult
                elif isinstance(client_message.root, types.ListResourcesRequest):
                    result_type = types.ListResourcesResult
                elif isinstance(client_message.root, types.ListResourceTemplatesRequest):
                    result_type = types.ListResourceTemplatesResult
                elif isinstance(client_message.root, types.ReadResourceRequest):
                    result_type = types.ReadResourceResult
                elif isinstance(client_message.root, types.SubscribeRequest):
                    result_type = types.EmptyResult
                elif isinstance(client_message.root, types.UnsubscribeRequest):
                    result_type = types.EmptyResult
                elif isinstance(client_message.root, types.ListPromptsRequest):
                    result_type = types.ListPromptsResult
                elif isinstance(client_message.root, types.GetPromptRequest):
                    result_type = types.GetPromptResult
                elif isinstance(client_message.root, types.CompleteRequest):
                    result_type = types.CompleteResult
                elif isinstance(client_message.root, types.ListToolsRequest):
                    result_type = types.ListToolsResult
                elif isinstance(client_message.root, types.CallToolRequest):
                    result_type = types.CallToolResult
                else:
                    # this is a special case where we need to return the list of tools.
                    raise NotImplementedError(f"Method {client_message.method} not implemented")
                # Call the mcp server using send_request() or send_notification() depending on the method.
                try:
                    return await self.client.session.send_request(client_message, result_type)
                except McpError as e:
                    return types.JSONRPCError(jsonrpc="2.0", id=id, error=e.error)

        # The message coming here is the generic request. We look at it's .method
        # to determine which client function to call and to further subparse the params.
        # Note(zeiler): unfortunately the pydantic types in mcp/types.py are not consistent.
        # The JSONRPCRequest are supposed to have an id but the InitializeRequest
        # does not have it.
        d = json.loads(msg)
        id = d.get('id', "")

        # If we have an id it's a JSONRPCRequest
        if not d.get('method', '').startswith("notifications/"):
            client_message = types.ClientRequest.model_validate(d)
            # Note(zeiler): this response is the "result" field of the JSONRPCResponse.
            # the API will fill in the "id" and "jsonrpc" fields.
            response = asyncio.run(send_request(client_message, id=id))
            if response is None:
                response = types.JSONRPCError(
                    jsonrpc="2.0",
                    id=id,
                    error=types.ErrorData(
                        code=types.INTERNAL_ERROR, message="Got empty response from MCP server."
                    ),
                )
            # return as a serialized json string
            res = response.model_dump_json(by_alias=True, exclude_none=True)
            return res
        else:  # JSONRPCRequest
            client_message = types.ClientNotification.model_validate(d)
            # send_notification returns None always so nothing to return.
            asyncio.run(send_notification(client_message))
            return "{}"



import itertools
import time
from concurrent.futures import ThreadPoolExecutor

import grpc
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.client.auth.register import RpcCallable, V2Stub
from clarifai.utils.logging import logger

throttle_status_codes = {
    status_code_pb2.CONN_THROTTLED,
    status_code_pb2.CONN_EXCEED_HOURLY_LIMIT,
}

retry_codes_grpc = {
    grpc.StatusCode.UNAVAILABLE,
}

_threadpool = ThreadPoolExecutor(100)


def validate_response(response, attempt, max_attempts):
    # Helper function to handle simple response validation
    def handle_simple_response(response):
        if hasattr(response, 'status') and hasattr(response.status, 'code'):
            if (response.status.code in throttle_status_codes) and attempt < max_attempts:
                logger.debug('Retrying with status %s' % str(response.status))
                return None  # Indicates a retry is needed
            else:
                return response

    # Check if the response is an instance of a gRPC streaming call
    if isinstance(response, grpc._channel._MultiThreadedRendezvous):
        try:
            # Check just the first response in the stream for validation
            first_res = next(response)
            validated_response = handle_simple_response(first_res)
            if validated_response is not None:
                # Have to return that first response and the rest of the stream.
                return itertools.chain([validated_response], response)
            return None  # Indicates a retry is needed
        except grpc.RpcError as e:
            logger.error('Error processing streaming response: %s' % str(e))
            return None  # Indicates an error
    else:
        # Handle simple response validation
        return handle_simple_response(response)


def create_stub(auth_helper: ClarifaiAuthHelper = None, max_retry_attempts: int = 10) -> V2Stub:
    """
    Create client stub that handles authorization and basic retries for
    unavailable or throttled connections.

    Args:
      auth_helper:  ClarifaiAuthHelper to use for auth metadata (default: from env)
      max_retry_attempts:  max attempts to retry rpcs with retryable failures
    """
    stub = AuthorizedStub(auth_helper)
    if max_retry_attempts > 0:
        return RetryStub(stub, max_retry_attempts)
    return stub


class AuthorizedStub(V2Stub):
    """V2Stub proxy that inserts metadata authorization in rpc calls."""

    def __init__(self, auth_helper: ClarifaiAuthHelper = None):
        if auth_helper is None:
            auth_helper = ClarifaiAuthHelper.from_env()
        self.stub = auth_helper.get_stub()
        self.metadata = auth_helper.metadata

    def __getattr__(self, name):
        value = getattr(self.stub, name)
        if isinstance(value, RpcCallable):
            value = _AuthorizedRpcCallable(value, self.metadata)
        return value


class _AuthorizedRpcCallable(RpcCallable):
    """Adds metadata(authorization header) to rpc calls"""

    def __init__(self, func, metadata):
        self.f = func
        self.metadata = metadata

    def __repr__(self):
        return repr(self.f)

    def __call__(self, *args, **kwargs):
        metadata = kwargs.pop('metadata', self.metadata)
        return self.f(*args, **kwargs, metadata=metadata)

    def future(self, *args, **kwargs):
        metadata = kwargs.pop('metadata', self.metadata)
        return self.f.future(*args, **kwargs, metadata=metadata)

    def __getattr__(self, name):
        return getattr(self.f, name)


class RetryStub(V2Stub):
    """
    V2Stub proxy that retries requests (currently on unavailable server or throttle codes)
    """

    def __init__(self, stub, max_attempts=10, backoff_time=5):
        self.stub = stub
        self.max_attempts = max_attempts
        self.backoff_time = backoff_time

    def __getattr__(self, name):
        value = getattr(self.stub, name)
        if isinstance(value, RpcCallable):
            value = _RetryRpcCallable(value, self.max_attempts, self.backoff_time)
        return value


class _RetryRpcCallable(RpcCallable):
    """Retries rpc calls on unavailable server or throttle codes"""

    def __init__(self, func, max_attempts, backoff_time):
        self.f = func
        self.max_attempts = max_attempts
        self.backoff_time = backoff_time

    def __repr__(self):
        return repr(self.f)

    def __call__(self, *args, **kwargs):
        attempt = 0
        while attempt < self.max_attempts:
            attempt += 1
            if attempt != 1:
                time.sleep(self.backoff_time)  # TODO better backoff between attempts
            try:
                response = self.f(*args, **kwargs)
                v = validate_response(response, attempt, self.max_attempts)
                if v is not None:
                    return v
            except grpc.RpcError as e:
                if (e.code() in retry_codes_grpc) and attempt < self.max_attempts:
                    logger.debug('Retrying with status %s' % e.code())
                else:
                    raise

    def future(self, *args, **kwargs):
        # TODO use single result event loop thread with asyncio
        return _threadpool.submit(self, *args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.f, name)
